/*
 *  linux/arch/arm/mach-psionw/irq.c
 *
 *  Copyright (C) 2001 Yuji Shinokawa <ysh@mob.or.jp>
 *  Copyright (C) 2001 Tony Lindgren <tony@atomide.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/init.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>

#include <linux/delay.h>
#include <linux/sched.h>

#include <asm/hardware/ioc.h>

#include <asm/hardware/psionw.h>

static int debug_on;
static struct irqaction *etna_action = NULL;

#if CONFIG_PCMCIA_ETNA
extern void etna_interrupt(void);

static struct tq_struct etna_int_task = {
	routine:etna_interrupt
};
#endif

static void
unmask_irq_int(unsigned int irq);

static void
mask_irq_int(unsigned int irq)
{
	u32 intmr;
	intmr = (1 << irq);
	psionw_writel(intmr, INTENC);
}

#ifdef CONFIG_PCMCIA_ETNA
static void
mask_etna_irq(unsigned int irq)
{
	int etna_int, count = 0;
	mask_irq_int(irq);

	etna_int = etna_readb(ETNA_INT_STATUS);
	etna_action = irq_desc[irq].action;

	//printk("i0x%x ", etna_int);

	switch (etna_int) {
	case (0x00):		/* Card removed interrupt */
		irq_desc[irq].action = NULL;
		printk("ETNA: Card removed\n");
		schedule_task(&etna_int_task);
		break;
	case (0x01):		/* Card got hosed */
		irq_desc[irq].action = NULL;
		while (etna_int == 0x01) {
			count++;
			schedule_task(&etna_int_task);
			mdelay(500);
			etna_int = etna_readb(ETNA_INT_STATUS);
			if (count >= 5) {
				printk("ETNA: Socket stuck, reinsert the card\n");
				break;
			}
		}
		break;
	case (0x03):
		break;		/* Network after nfs mount */
	case (0x05):		/* Normal IDE interrupt */
		break;
	case (0x07):		/* Normal network interrupt */
		break;
	case (0x3f):		/* First write after sleep mode */
	case (0x35):		/* First write after sleep mode */
		//irq_desc[IRQ_EINT1].action = NULL;
		break;
	default:
		printk("ETNA: Unknown interrupt status: 0x%x\n",  etna_int);
		schedule_task(&etna_int_task);
	}

	/* 
	 * You would assume that we need to disable Etna interrupt 
	 * until we're done? Nope, that causes delays in waking up
	 * the card for 16-bit writes in etna_outsw_ide.
	 */
	//etna_writeb(0, ETNA_INT_MASK);

	/* Clear the Etna interrupt */
	etna_writeb(ETNA_CLEAR_MASK, ETNA_INT_CLEAR);
}

static void
unmask_etna_irq(unsigned int irq)
{
	int etna_int_mask;

	etna_int_mask = etna_readb(ETNA_INT_MASK);
	while (etna_int_mask & (1 << 2)) {
		printk("ETNA: Bad ETNA_INT_MASK: 0x%02x\n", 
		       etna_readb(ETNA_INT_STATUS));
		etna_writeb(0, ETNA_INT_MASK);
		etna_writeb(ETNA_CF_IRQ, ETNA_INT_CLEAR);
		etna_int_mask = etna_readb(ETNA_INT_MASK);
	}

	if (irq_desc[irq].action == NULL)
		irq_desc[irq].action = etna_action;

	/* Let's enable the Etna interrupt again */
	etna_writeb(ETNA_CF_IRQ, ETNA_INT_MASK);
	unmask_irq_int(irq);
}
#endif /* CONFIG_PCMCIA_ETNA */

static void
mask_ack_irq_int(unsigned int irq)
{
	switch (irq) {
	case IRQ_CSINT:
		psionw_writel(1, COEOI);
		printk("Received and cleared CSINT irq %d\n", irq);
		break;
	case IRQ_EINT2:
		psionw_writel(1, E2EOI);
		printk("Received and cleared EINT2 irq %d\n", irq);
		break;
	case IRQ_TC1OI:
		psionw_writel(1, TC1EOI); /* Cleared here, handled in rtc */
		break;
	case IRQ_TC2OI:
		psionw_writel(1, TC2EOI); /* Cleared here, handled in timer */
		break;
	case IRQ_RTCMI:
		psionw_writel(1, RTCEOI); /* Cleared here, handled in rtc*/
		break;
	case IRQ_TINT:
		psionw_writel(1, TEOI);
		printk("Received and cleared TINT irq %d\n", irq);
		break;
	}
}

static void
unmask_irq_int(unsigned int irq)
{
	u32 intmr;
	intmr = (1 << irq);
	psionw_writel(intmr, INTENS);
}

void __init
psionw_init_irq(void)
{
	unsigned int i;

	for (i = 0; i < NR_IRQS; i++) {
		if (INT1_IRQS & (1 << i)) {
			irq_desc[i].valid = 1;
			irq_desc[i].probe_ok = 1;
			irq_desc[i].mask_ack = (INT1_ACK_IRQS & (1 << i)) ?
			    mask_ack_irq_int : mask_irq_int;
			irq_desc[i].mask = mask_irq_int;
			irq_desc[i].unmask = unmask_irq_int;
		}
	}

	/*
	 * Disable interrupts
	 */
	psionw_writel(~0, INTENC);

	/*
	 * Clear down any pending interrupts
	 */
	psionw_writel(0, MCEOI);
	psionw_writel(0, BLEOI);
	psionw_writel(0, COEOI);
	psionw_writel(0, TC1EOI);
	psionw_writel(0, TC2EOI);
	psionw_writel(0, RTCEOI);
	psionw_writel(0, TEOI);
	psionw_writel(0, UMSEOI);
	psionw_writel(0, SSCR0);

#ifdef CONFIG_PCMCIA_ETNA
	irq_desc[IRQ_EINT1].mask_ack = mask_etna_irq;
	irq_desc[IRQ_EINT1].unmask = unmask_etna_irq;

	/* Enable the ETNA interrupts */
	__raw_writeb(ETNA_CF_IRQ, ETNA_V_BASE + ETNA_INT_MASK);
#endif

	init_FIQ();
}
